# Copyright (c) 2013, Sascha Peilicke <saschpe@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (see the file COPYING); if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
Open Build Service plugin.
"""

import lxml.etree
import requests

import rapport.plugin


class OpenBuildServicePlugin(rapport.plugin.Plugin):
    def __init__(self, *args, **kwargs):
        super(OpenBuildServicePlugin, self).__init__(*args, **kwargs)

    def _get_xml(self, url):
        response = requests.get(url, auth=(self.login, self.password))
        return lxml.etree.fromstring(response.text)

    def collect(self, timeframe):
        # Directly querying for all user's requests is killing the OBS api.
        # So we ask for all projects / packages where the user is involved in first.

        user_projects_url = "{0}/search/project/id?match=person/@userid='{1}'".format(self.url.geturl(), self.login)
        xml_root = self._get_xml(user_projects_url)
        user_projects = [project.get('name') for project in xml_root.iterfind(".//project")]

        #user_packages_url = "{0}/search/package/id?match=person/@userid='{1}'".format(self.url.geturl(), self.login)
        #xml_root = self._get_xml(user_projects_url)
        #user_packages = [package.values() for package in xml_root.iterfind(".//package")]

        #"https://api.opensuse.org/search/request/id?match=history/@who='saschpe'"

        for project in user_projects:
            requests = self._get_xml("{0}/request?view=collection&user={1}&project={2}".format(self.url.geturl(), self.login, project))


rapport.plugin.register("openbuildservice", OpenBuildServicePlugin)



# Copyright (c) 2013, Sascha Peilicke <saschpe@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (see the file COPYING); if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""Various utility functions.
"""

import datetime
import re
import subprocess

_FIRST_CAP_RE = re.compile('(.)([A-Z][a-z]+)')
_ALL_CAP_RE = re.compile('([a-z0-9])([A-Z])')


def camelcase_to_underscores(word):
    """Converts a CamelCase word into an under_score word.

        >>> camelcase_to_underscores("CamelCaseCase")
        'camel_case_case'
        >>> camelcase_to_underscores("getHTTPResponseCode")
        'get_http_response_code'
    """
    s1 = _FIRST_CAP_RE.sub(r'\1_\2', word)
    return _ALL_CAP_RE.sub(r'\1_\2', s1).lower()


def silent_popen(args, **kwargs):
    """Wrapper for subprocess.Popen with suppressed output.

    STERR is redirected to STDOUT which is piped back to the
    calling process and returned as the result.
    """
    return subprocess.Popen(args,
                            stderr=subprocess.STDOUT,
                            stdout=subprocess.PIPE, **kwargs).communicate()[0]


def datetime_from_iso8601(date):
    """Small helper that parses ISO-8601 date dates.

        >>> datetime_from_iso8601("2013-04-10T12:52:39")
        datetime.datetime(2013, 4, 10, 12, 52, 39)
        >>> datetime_from_iso8601("2013-01-07T12:55:19.257")
        datetime.datetime(2013, 1, 7, 12, 55, 19, 257000)
    """
    format = "%Y-%m-%dT%H:%M:%S"
    if date.endswith("Z"):
        date = date[:-1]  # Date date is UTC
    if re.match(".*\.\d+", date):
        format += ".%f"  # Date includes microseconds
    return datetime.datetime.strptime(date, format)



# Copyright (c) 2013, Sascha Peilicke <saschpe@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (see the file COPYING); if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import glob
import re
import shutil
import subprocess
import sys
from distutils.core import Command


class PEP8Command(Command):
    description = "Run PEP8 with custom options"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        subprocess.call(["pep8", "--repeat", "--show-source",
                        "--ignore=E501",
                        "--exclude=.venv,.tox,dist,doc", "."])


class CleanupCommand(Command):
    patterns = [".tox", ".venv", "build", "dist", "*.egg-info"]
    description = "Clean up project directory"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        for pattern in CleanupCommand.patterns:
            for file in glob.glob(pattern):
                shutil.rmtree(file, ignore_errors=True)


def get_cmdclass():
    """Dictionary of all distutils commands defined in this module.
    """
    return {"cleanup": CleanupCommand,
            "pep8": PEP8Command}


def parse_requirements(requirements_file='requirements.txt'):
    requirements = []
    with open(requirements_file, 'r') as f:
        for line in f:
            # For the requirements list, we need to inject only the portion
            # after egg= so that distutils knows the package it's looking for
            # such as:
            # -e git://github.com/openstack/nova/master#egg=nova
            if re.match(r'\s*-e\s+', line):
                requirements.append(re.sub(r'\s*-e\s+.*#egg=(.*)$', r'\1',
                                    line))
            # such as:
            # http://github.com/openstack/nova/zipball/master#egg=nova
            elif re.match(r'\s*https?:', line):
                requirements.append(re.sub(r'\s*https?:.*#egg=(.*)$', r'\1',
                                    line))
            # -f lines are for index locations, and don't get used here
            elif re.match(r'\s*-f\s+', line):
                pass
            # -r lines are for including other files, and don't get used here
            elif re.match(r'\s*-r\s+', line):
                pass
            # argparse is part of the standard library starting with 2.7
            # adding it to the requirements list screws distro installs
            elif line == 'argparse' and sys.version_info >= (2, 7):
                pass
            else:
                requirements.append(line.strip())
    return requirements



# Copyright (c) 2013, Sascha Peilicke <saschpe@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (see the file COPYING); if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
Command-line interface to rapport.
"""

import argparse
import getpass
import os
import sys
if sys.version_info > (3, 3):
    import concurrent.futures as futures
else:
    import futures

# Custom hack for running rapport/cli inside the development tree:
if __name__ == "__main__":
    sys.path.append(os.getcwd())

import rapport
import rapport.config
import rapport.plugin
import rapport.template
import rapport.timeframe


rapport.config.load()
rapport.config.init_user()
rapport.plugin.discover()


class CLI(object):
    def __init__(self):
        self.plugins = rapport.plugin.init_from_config()
        self.timeframe = rapport.timeframe.init_from_config()

    def create(self, args):
        results = {}
        with futures.ThreadPoolExecutor(max_workers=4) as executor:
            plugin_futures = dict((executor.submit(p.collect, self.timeframe), p) for p in self.plugins)
            for future in futures.as_completed(plugin_futures):
                plugin = plugin_futures[future]
                try:
                    if rapport.config.get_int("rapport", "verbosity") >= 2:
                        print "Result for {0}: {1}".format(plugin.alias, future.result())
                    template = rapport.template.get_template(plugin, "text")
                    if template:
                        results[plugin] = template.render(future.result())
                except Exception as e:
                    print >>sys.stderr, "Failed plugin {0}:{1}: {2}!".format(plugin, plugin.alias, e)

        # TODO: Generate mail template
        # TODO: refactor this into appropriate modules:
        template_email_subject = rapport.template.get_template("subject", type="email")
        template_email_body = rapport.template.get_template("body", type="email")
        #email_subject = template_email_subject.render(
        email_body = template_email_body.render({"plugins": self.plugins, "results": results})
        
        report_path = os.path.expanduser(os.path.join("~", ".rapport", "reports", self.timeframe.end.isoformat()))
        if not os.path.exists(report_path):
            os.makedirs(report_path)
        report_file = os.path.join(report_path, "email.body.text")
        with open(report_file, "w") as report:
            report.write(email_body)

        print "Your work report ({0}):\n{1}".format(report_file, email_body)

       ## Print results sorted by plugin appearance in config file (i.e. init order):
       #for plugin in self.plugins:
       #    try:
       #        print results[plugin]
       #    except KeyError as e:
       #        # A missing result for plugins means an exception happened
       #        # above, which already printed an error message, thus:
       #        pass


   #def edit(self, args):
   #    pass

    def main(self):
        parser = argparse.ArgumentParser()
        parser.add_argument("--version", action="version", version="rapport {0}".format(rapport.__version__))
        subparsers = parser.add_subparsers(title="commands")

        parser_create = subparsers.add_parser("create", help="create work report")
        parser_create.set_defaults(func=self.create)

       #parser_edit = subparsers.add_parser("edit", help="edit report prior to sending")
       #parser_edit.set_defaults(func=self.edit)

        parser_help = subparsers.add_parser("help", help="show this help")
        parser_help.set_defaults(func=lambda args: parser.print_help())

        args = parser.parse_args()
        args.func(args)


def main():
   #try:
        CLI().main()
   #except Exception as e:
   #    print >> sys.stderr, e
   #    sys.exit(1)


if __name__ == "__main__":
    sys.exit(main())



# Copyright (c) 2013, Sascha Peilicke <saschpe@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (see the file COPYING); if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import datetime

import rapport.config
import rapport.util


class Timeframe(object):
    """Represents a period of time between a start and end time.

    :start: Start of timeframe (datetime object)
    :end: End of timeframe (datetime object)
    """
    def __init__(self, start, end):
        self._start = start
        self._end = end

    @property
    def start(self):
        return self._start

    @property
    def end(self):
        return self._end

    def contains(self, date):
        """Checks if a date is within a timeframe.

        :date: The date to check
        """
        return self._start <= date and date < self._end

    def __str__(self):
        """Returns the class name in underscores.

        Additionally, for sub-classes, the suffix '_timeframe' is split off.

            >>> from rapport import timeframe
            >>> t = timeframe.Timeframe(start=None, end=None)
            >>> str(t)
            'timeframe'
        """
        return rapport.util.camelcase_to_underscores(self.__class__.__name__).rsplit("_timeframe")[0]


class CurrentWeekTimeframe(Timeframe):
    """Current week timeframe (in UTC).
    """
    def __init__(self):
        self._end = datetime.datetime.utcnow()
        # Compute the day but reset the hours/minutes/seconds to zero,
        # we want the exact week's start:
        week_start = self._end - datetime.timedelta(days=self._end.weekday())
        self._start = datetime.datetime(year=week_start.year,
                                        month=week_start.month,
                                        day=week_start.day)


class WeekTimeframe(Timeframe):
    """N-th week of year timeframe (in UTC).

    :week: Week number (starting from 1)
    """
    def __init__(self, week=1):
        raise NotImplementedError()


class CurrentMonthTimeframe(Timeframe):
    """Current month timeframe (in UTC).
    """
    def __init__(self):
        self._end = datetime.datetime.utcnow()
        self._start = datetime.datetime(year=self._end.year,
                                        month=self._end.month, day=1)


class MonthTimeframe(Timeframe):
    """N-th month of year timeframe (in UTC).

    :month: Month number (starting from 1)
    """
    def __init__(self, month=1):
        raise NotImplementedError()


class NLastDaysTimeframe(Timeframe):
    """'N' last days timeframe (in UTC).
    """
    def __init__(self, days=7):
        self._end = datetime.datetime.utcnow()
        self._start = self._end - datetime.timedelta(days=days)


_TIMEFRAME_CATALOG = {"current_month": CurrentMonthTimeframe,
                      "current_week": CurrentWeekTimeframe,
                      "month": MonthTimeframe,
                      "week": WeekTimeframe,
                      "n_last_days": NLastDaysTimeframe}


def init(name, *args, **kwargs):
    """Instantiate a timeframe from the catalog.
    """
    if name in _TIMEFRAME_CATALOG:
        if rapport.config.get_int("rapport", "verbosity") >= 2:
            print "Initialize timeframe {0}: {1} {2}".format(name, args, kwargs)
        try:
            return _TIMEFRAME_CATALOG[name](*args, **kwargs)
        except ValueError as e:
            print >>sys.stderr, "Failed to initialize timeframe {0}: {1}!".format(name, e)
    else:
        print >>sys.stderr, "Failed to initialize timeframe {0}: Not in catalog!".format(name)


def init_from_config():
    return init(rapport.config.get("timeframe", "default"))


def catalog():
    """Returns the list of registered timeframes.
    """
    return _TIMEFRAME_CATALOG.keys()



# Copyright (c) 2013, Sascha Peilicke <saschpe@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (see the file COPYING); if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
Mediawiki plugin.
"""

import json

import requests

import rapport.plugin


class MediawikiPlugin(rapport.plugin.Plugin):
    def __init__(self, *args, **kwargs):
        super(MediawikiPlugin, self).__init__(*args, **kwargs)

    def _get(self, params={}):
        params.update({"format": "json", "action": "query"})
        response = requests.get(self.url.geturl(), params=params, auth=(self.login, self.password))
        return json.loads(response.text)["query"]

    def collect(self, timeframe):
        #TODO: Check back ugly "UTC" timestamp hack here:
        edits = self._get({"list": "usercontribs",
                                   "ucuser": "{0}".format(self.login),
                                   "ucstart": timeframe.end.isoformat()+"Z",
                                   "ucend": timeframe.start.isoformat()+"Z"})
        return self._results({"edits": edits["usercontribs"]})


rapport.plugin.register("mediawiki", MediawikiPlugin)



#
# Copyright (c) 2013, Sascha Peilicke <saschpe@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (see the file COPYING); if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
"""
:mod:`rapport` -- Work Report Generator
=======================================

.. automodule:: rapport
   :synopsis: Work Report Generator
.. moduleauthor:: Sascha Peilicke <saschpe@gmx.de>
"""

__author__ = "Sascha Peilicke <saschpe@gmx.de>"
__version_info__ = ("0", "0", "8")
__version__ = ".".join(__version_info__)



# Copyright (c) 2013, Sascha Peilicke <saschpe@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (see the file COPYING); if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import os
import site
import sys

import jinja2


def _get_template_dirs(type="plugin"):
    """Return a list of directories where templates may be located.
    """
    template_dirs = [
        os.path.expanduser(os.path.join("~", ".rapport", "templates", type)),
        os.path.join("rapport", "templates", type)  # Local dev tree
    ]  + map(lambda d: os.path.join(d, "rapport", "templates", type), site.getsitepackages())
    return template_dirs


_JINJA2_ENV = {}
def init():
    for type in ["plugin", "email", "web"]:
        loader = jinja2.FileSystemLoader(_get_template_dirs(type))
        env = jinja2.Environment(loader=loader,
                                 extensions=["jinja2.ext.i18n"],
                                 trim_blocks=True)
        env.install_null_translations(newstyle=False)
        _JINJA2_ENV[type] = env


def get_template(name, format="text", type="plugin"):
    if not _JINJA2_ENV:
        init()
    template_name = "{0}.{1}.jinja2".format(name, format)
    try:
        return _JINJA2_ENV[type].get_template(template_name)
    except jinja2.TemplateNotFound as e:
        print >>sys.stderr, "Missing template {0}/{1}!".format(type, template_name)




# Copyright (c) 2013, Sascha Peilicke <saschpe@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (see the file COPYING); if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
Bugzilla plugin.
"""

# Avoid name clash with global "bugzilla" module from python-bugzilla:
from __future__ import absolute_import
import bugzilla

import rapport.plugin


class BugzillaPlugin(rapport.plugin.Plugin):
    def __init__(self, email, *args, **kwargs):
        super(BugzillaPlugin, self).__init__(*args, **kwargs)
        self.email = email

        url = "{0}/xmlrpc.cgi".format(self.url.geturl())
        self._bz = bugzilla.Bugzilla(url=url)
        self._bz.login(user=self.login, password=self.password)

    def collect(self, timeframe):
        open_bugs = self._bz.query({"assigned_to": self.email,
                                    "bug_status": ["NEW", "ASSIGNED"]})
        closed_bugs = self._bz.query({"assigned_to": self.email,
                                      "bug_status": ["CLOSED", "RESOLVED"],
                                      "last_change_time": timeframe.start})

        return self._results({"email": self.email, "open_bugs": open_bugs,
                              "closed_bugs": closed_bugs})


rapport.plugin.register("bugzilla", BugzillaPlugin)



# Copyright (c) 2013, Sascha Peilicke <saschpe@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (see the file COPYING); if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
Launchpad plugin.
"""

import warnings

warnings.filterwarnings('ignore', 'Module argparse was already imported')   # Filter a UserWarning from Jinja2
from launchpadlib.launchpad import Launchpad

import rapport.plugin


class LaunchpadPlugin(rapport.plugin.Plugin):
    def __init__(self, *args, **kwargs):
        super(LaunchpadPlugin, self).__init__(*args, **kwargs)

        self.lp = Launchpad.login_anonymously(self.login, 'production')

    def _get_json(url):
        return json.loads(requests.get(url).text)

    def collect(self, timeframe):
        bug_tasks = self.lp.people["saschpe"].searchTasks()

        #TODO: Try to find some useful info
        return self._results()


rapport.plugin.register("launchpad", LaunchpadPlugin)



# Copyright (c) 2013, Sascha Peilicke <saschpe@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (see the file COPYING); if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import os
import site
import sys
import traceback
import urlparse

import rapport.config
import rapport.util


class Plugin(object):
    def __init__(self, alias=None, url=None, login=None, password=None):
        self.alias = alias
        self.url = url
        self.login = login
        self.password = password

        if alias == "":
            self.alias == self.__str__()

        if url and type(url) is not urlparse.ParseResult:
            self.url = urlparse.urlparse(url)

    def _results(self, dict={}):
        """Helper to merge a dict with cross-plugin defaults.

        All plugin sub-classes share some config values, i.e.  alias, url,
        login and password. This help should be used in the :collect: method
        of any Plugin implementation.

           >>> import rapport.plugin
           >>> c = rapport.plugin.Plugin(alias="a", url="http://example.com", login="u")
           >>> c._results()
           {'url': 'http://example.com', 'alias': 'a', 'login': 'u'}
           >>> c._results({'mykey': 'mykey'})
           {'url': 'http://example.com', 'alias': 'a', 'login': 'u', 'mykey': 'mykey'}
        """
        results = {"plugin": str(self),
                   "alias": self.alias,
                   "url": self.url,
                   "login": self.login}
        results.update(dict)
        return results

    def __str__(self):
        """Returns the class name in underscores.

        Additionally, for sub-classes, the suffix '_plugin' is split off.

            >>> import rapport.plugin
            >>> c = rapport.plugin.Plugin()
            >>> str(c)
            'plugin'
        """
        return rapport.util.camelcase_to_underscores(self.__class__.__name__).rsplit("_plugin")[0]

    def collect(self, timeframe):
        raise NotImplementedError()


def _get_plugin_dirs():
    """Return a list of directories where plugins may be located.
    """
    plugin_dirs = [
        os.path.expanduser(os.path.join("~", ".rapport", "plugins")),
        os.path.join("rapport", "plugins")  # Local dev tree
    ]  + map(lambda d: os.path.join(d, "rapport", "plugins"), site.getsitepackages())
    return plugin_dirs


def _path_to_module(path):
    """Translates paths to *.py? files into module paths.

        >>> _path_to_module("rapport/bar.py")
        'rapport.bar'
        >>> _path_to_module("/usr/lib/rapport/bar.py")
        'rapport.bar'
    """
    # Split of preceeding path elements:
    path = "rapport" + path.split("rapport")[1]
    # Split of ending and replace os.sep with dots:
    path = path.replace(os.sep, ".").rsplit(".", 1)[0]
    return path


def discover():
    """Find and load all available plugins.
    """
    plugin_files = []

    for plugin_dir in _get_plugin_dirs():
        if os.path.isdir(plugin_dir):
            for plugin_file in os.listdir(plugin_dir):
                if plugin_file.endswith(".py") and not plugin_file == "__init__.py":
                    plugin_files.append(os.path.join(plugin_dir, plugin_file))

    if rapport.config.get_int("rapport", "verbosity") >= 2:
        print "Found plugin modules: {0}".format(plugin_files)
    for plugin_file in plugin_files:
        if rapport.config.get_int("rapport", "verbosity") >= 2:
            print "Importing module {0}".format(_path_to_module(plugin_file))
        __import__(_path_to_module(plugin_file))


_PLUGIN_CATALOG = {}


def register(name, klass):
    """Add a plugin to the plugin catalog.
    """
    if rapport.config.get_int("rapport", "verbosity") >= 1:
        print "Registered plugin: {0}".format(name)
    _PLUGIN_CATALOG[name] = klass


def init(name, *args, **kwargs):
    """Instantiate a plugin from the catalog.
    """
    if name in _PLUGIN_CATALOG:
        if rapport.config.get_int("rapport", "verbosity") >= 2:
            print "Initialize plugin {0}: {1} {2}".format(name, args, kwargs)
        try:
            return _PLUGIN_CATALOG[name](*args, **kwargs)
        except (ValueError, TypeError) as e:
            print >>sys.stderr, "Failed to initialize plugin {0}: {1}!".format(name, e)
    else:
        print >>sys.stderr, "Failed to initialize plugin {0}: Not in catalog!".format(name)


def init_from_config():
    plugins = []
    for plugin in rapport.config.plugins():
        plugins.append(init(**plugin))
    return filter(bool, plugins)


def catalog():
    """Returns the list of registered plugins.
    """
    return _PLUGIN_CATALOG.keys()



# Copyright (c) 2013, Sascha Peilicke <saschpe@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (see the file COPYING); if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
Twitter plugin.
"""

import tweepy

import rapport.plugin


class TwitterPlugin(rapport.plugin.Plugin):
    def __init__(self, consumer_key, consumer_secret, access_token, access_token_secret, *args, **kwargs):
        super(TwitterPlugin, self).__init__(*args, **kwargs)

        auth = tweepy.OAuthHandler(consumer_key, consumer_secret)
        auth.set_access_token(access_token, access_token_secret)

        self.api = tweepy.API(auth)

    def _get(self, timeline, timeframe):
        #TODO: Implement proper pagination instead of fixed count:
        count = 50
        statuses = []
        for status in timeline(count=count):
            if timeframe.contains(status.created_at):
                statuses.append(status)
        return statuses

    def collect(self, timeframe):
        mentions = self._get(self.api.mentions_timeline, timeframe)
        tweets = self._get(self.api.user_timeline, timeframe)
        retweets = self._get(self.api.retweets_of_me, timeframe)

        return self._results({"mentions": mentions,
                              "tweets": tweets,
                              "retweets": retweets})


rapport.plugin.register("twitter", TwitterPlugin)



# Copyright (c) 2013, Sascha Peilicke <saschpe@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (see the file COPYING); if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""
Gerrit plugin.
"""

import json
from datetime import datetime

import paramiko

import rapport.plugin


class GerritPlugin(rapport.plugin.Plugin):
    def __init__(self, *args, **kwargs):
        super(GerritPlugin, self).__init__(*args, **kwargs)

        self._client = paramiko.SSHClient()
        self._client.load_system_host_keys()

    def _ssh_cmd(self, *args):
        """Execute a gerrit command over SSH.
        """
        command = "gerrit {0}".format(" ".join(args))
        _, stdout, stderr = self._client.exec_command(command)
        return (stdout.readlines(), stderr.readlines())

    def _ssh_query(self, *args):
        """Execute a gerrit query over SSH and returns JSON-formatted data.
        """
        return self._ssh_cmd("query", "--format=JSON", *args)

    def collect(self, timeframe):
        self._client.connect(self.url.hostname, self.url.port, self.login)
        stdout, stderr = self._ssh_query("owner:{0}".format(self.login))
        self._client.close()

        changes = []
        if not stderr:
            for line in stdout[:-1]:  # Last line contains only download stats
                change = json.loads(line)
                if "lastUpdated" in change:
                    last_updated = datetime.utcfromtimestamp(
                        change["lastUpdated"])
                    if timeframe.contains(last_updated):
                        changes.append(change)
                else:
                    print("Change {0} is missing lastUpdated".format(change))

        return self._results({"changes": changes})


rapport.plugin.register("gerrit", GerritPlugin)



# Copyright (c) 2013, Sascha Peilicke <saschpe@gmx.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program (see the file COPYING); if not, write to the
# Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import os
import shutil
import ConfigParser

import rapport.config


def _get_config_dirs():
    """Return a list of directories where config files may be located.

    The following directories are returned::

      ~/.rapport/
      /etc/rapport/
    """
    config_dirs = [
        os.path.expanduser(os.path.join("~", ".rapport")),
        os.path.join("/", "etc", "rapport"),
        os.path.abspath("config")
    ]
    return config_dirs


def find_config_files():
    """Return a list of default configuration files.
    """
    config_files = []

    for config_dir in _get_config_dirs():
        path = os.path.join(config_dir, "rapport.conf")
        if os.path.exists(path):
            config_files.append(path)

    return filter(bool, config_files)


def _get_reports_dir():
    return os.path.expanduser(os.path.join("~", ".rapport", "reports")),


def init_user():
    """Create and populate the ~/.rapport directory tree if it's not existing.

    Doesn't interfere with already existing directories or configuration files.
    """
    user_conf_dir = os.path.expanduser(os.path.join("~", ".rapport"))
    user_conf_file = os.path.join(user_conf_dir, "rapport.conf")

    if not os.path.exists(user_conf_dir):
        if rapport.config.get_int("rapport", "verbosity") >= 1:
            print "Create user directory {0}".format(user_conf_dir)
        os.makedirs(user_conf_dir)
    for subdir in ["plugins", "reports", "templates/plugin", "templates/email", "templates/web"]:
        user_conf_subdir = os.path.join(user_conf_dir, subdir)
        if not os.path.exists(user_conf_subdir):
            if rapport.config.get_int("rapport", "verbosity") >= 1:
                print "Create user directory {0}".format(user_conf_subdir)
            os.makedirs(user_conf_subdir)
        if subdir == "reports" and not (os.stat(user_conf_subdir).st_mode & 0777) == 0700:
            if rapport.config.get_int("rapport", "verbosity") >= 1:
                print "Set secure directory permissions for {0}".format(user_conf_subdir)
            os.chmod(user_conf_subdir, 0700)
    if not os.path.exists(user_conf_file):
        if rapport.config.get_int("rapport", "verbosity") >= 1:
            print "Create user configuration {0}".format(user_conf_file)
        default_config = find_config_files()
        shutil.copyfile(default_config[0], user_conf_file)
    if not (os.stat(user_conf_file).st_mode & 0777) == 0600:
        if rapport.config.get_int("rapport", "verbosity") >= 1:
            print "Set secure file permissions for {0}".format(user_conf_file)
        os.chmod(user_conf_file, 0600)


CONF = None


def load():
    global CONF
    config = ConfigParser.SafeConfigParser()
    config.read(find_config_files()[0])
    CONF = config
    return CONF


def get(section, option):
    return CONF.get(section, option)


def get_int(section, option):
    return int(get(section, option))


def plugins():
    for section in CONF.sections():
        if section.startswith("plugin:"):
            name, alias = section.split(":")[1:]
            plugin = {"name": name, "alias": alias}
            for option in CONF.options(section):
                plugin[option] = CONF.get(section, option)
            yield plugin
